# Copyright (c) AppDynamics, Inc., and its affiliates
# 2015
# All Rights Reserved

"""Intercept psycopg2.

Note: this file also contains the old dbapi interceptor code.

"""

from functools import wraps
import re
import sys

from appdynamics.lib import parse_url, parse_qs


def make_connect_interceptor(agent, backend_naming_fn, connect_fn):
    @wraps(connect_fn)
    def connect(*args, **kwargs):
        try:
            backend = backend_naming_fn(*args, **kwargs)
        except:
            agent.logger.warning('could not name backend: %s', backend_naming_fn.__name__)
            return connect_fn(*args, **kwargs)

        with agent.db_exit_call(sys._getframe(1), backend, 'connect'):
            real_conn = connect_fn(*args, **kwargs)
            instrumented_conn = InstrumentedConnection(agent, backend, real_conn)
        return instrumented_conn

    connect.appd_orig_connect = connect_fn
    return connect


class InstrumentedConnection(object):
    def __init__(self, agent, backend, conn):
        super(InstrumentedConnection, self).__init__()
        self._appd_agent = agent
        self._appd_backend = backend
        self._appd_conn = conn

    def __getattribute__(self, name):
        if name in ('_appd_agent', '_appd_backend', '_appd_conn', 'cursor'):
            return object.__getattribute__(self, name)
        conn = object.__getattribute__(self, '_appd_conn')
        return getattr(conn, name)

    def cursor(self, *args, **kwargs):
        agent = object.__getattribute__(self, '_appd_agent')
        backend = object.__getattribute__(self, '_appd_backend')
        conn = object.__getattribute__(self, '_appd_conn')

        real_cursor = conn.cursor(*args, **kwargs)
        instrumented_cursor = InstrumentedCursor(agent, backend, real_cursor)
        return instrumented_cursor


class InstrumentedCursor(object):
    def __init__(self, agent, backend, cursor):
        super(InstrumentedCursor, self).__init__()
        self._appd_agent = agent
        self._appd_backend = backend
        self._appd_cursor = cursor

    def __getattribute__(self, name):
        if name in ('_appd_agent', '_appd_backend', '_appd_cursor', 'execute', '__iter__'):
            return object.__getattribute__(self, name)
        cursor = object.__getattribute__(self, '_appd_cursor')
        return getattr(cursor, name)

    def __iter__(self):
        cursor = object.__getattribute__(self, '_appd_cursor')
        return iter(cursor)

    def execute(self, operation, parameters=None):
        agent = object.__getattribute__(self, '_appd_agent')
        backend = object.__getattribute__(self, '_appd_backend')
        cursor = object.__getattribute__(self, '_appd_cursor')

        with agent.db_exit_call(sys._getframe(1), backend, operation, parameters):
            return cursor.execute(operation, parameters)


VENDOR = 'POSTGRESQL'
KV_RE = re.compile('\s*(?P<key>[a-zA-Z_]+)\s*=\s*(?P<remainder>.+)')


# For format of PostgreSQL connection strings, see:
# http://www.postgresql.org/docs/current/static/libpq-connect.html#LIBPQ-CONNSTRING


def parse_postgresql_url_dsn(dsn):
    url = parse_url(dsn, allow_fragments=False)

    if '?' in url.path:
        parts = url.path.split('?', 1)
        path, params = parts[0], parse_qs(parts[1])
    else:
        path = url.path
        params = parse_qs(url.query) if url.query else {}

    if 'host' in params:
        host = params['host'][0]
    else:
        host = url.hostname or 'localhost'

    if 'port' in params:
        port = params['port'][0]
    elif url.port:
        port = str(url.port)
    else:
        port = '5432'

    path = path.strip('/')

    if 'dbname' in params:
        dbname = params['dbname'][0]
    elif path:
        dbname = path
    elif 'user' in params:
        dbname = params['user'][0]
    else:
        dbname = url.username or 'postgres'

    return VENDOR, host, port, dbname


def parse_quoted_value(remainder):
    length = len(remainder)
    idx = 1

    start_idx = idx
    val = []

    while idx < length and remainder[idx] != "'":
        if remainder[idx:idx+2] in ("\\\\", "\\'"):
            # Valid escape sequence: collect the current run with the
            # escaped character, move past the escape sequence, and start a
            # new run.
            val.append(remainder[start_idx:idx] + remainder[idx + 1])
            idx += 2
            start_idx = idx
        else:
            idx += 1

    val.append(remainder[start_idx:idx])
    return ''.join(val), remainder[idx + 1:]


def parse_postgresql_kv_dsn(dsn):
    if "'" not in dsn:  # Fast path when there are no quoted strings.
        dsn = re.sub('(\s*=\s+|\s+=\s*)', '=', dsn).split()
        parts = dict(pair.split('=', 1) for pair in dsn)
    else:
        parts = {}
        m = KV_RE.match(dsn)

        while m:
            key, remainder = m.group('key'), m.group('remainder')

            if remainder[0] == "'":
                value, remainder = parse_quoted_value(remainder)
            else:
                split = remainder.split(None, 1)
                if len(split) == 1:
                    value, remainder = split[0], ''
                elif split:
                    value, remainder = split
                else:
                    break

            parts[key] = value
            m = KV_RE.match(remainder)

    return parse_postgresql_keyword_args(parts)


def parse_postgresql_keyword_args(parts):
    host = parts.get('host', 'localhost')
    port = parts.get('port', '5432')
    dbname = parts.get('dbname', parts.get('user', 'postgres'))
    return VENDOR, host, port, dbname


def parse_postgresql_dsn(dsn=None, *args, **kwargs):
    if dsn:
        if dsn.startswith('postgresql://'):
            return parse_postgresql_url_dsn(dsn)
        else:
            return parse_postgresql_kv_dsn(dsn)
    else:
        return parse_postgresql_keyword_args(kwargs)


def intercept_psycopg2(agent, mod):
    original_register_type = mod._ext.register_type

    def register_type(type, obj=None):
        if obj:
            try:
                obj = obj._appd_conn
            except:
                try:
                    obj = obj._appd_cursor
                except:
                    pass
        original_register_type(type, obj)
    mod._ext.register_type = register_type

    mod.connect = make_connect_interceptor(agent, parse_postgresql_dsn, mod.connect)
